
# THIS FILE IS GENERATED FROM MDTRAJ SETUP.PY
short_version = '0.4.1'
version = '0.4.1'
full_version = '0.4.1'
git_revision = '7e4b56234e284ba49f0b6819f6063ea6b5ec004a'
release = True

if not release:
    version = full_version
